@@ -2,6 +2,8 @@
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
+#include <ctype.h>
 #include <avr/io.h>
 #include <avr/wdt.h>
 #include <avr/power.h>
@@ -21,8 +23,8 @@ uint8_t CPS = 1<<PA6; // Crosspoint Strobe
 uint8_t CPD = 1<<PA7; // Crosspoint Data
 uint8_t CPR = 1<<PD7; // Crosspoint Reset
 
-uint8_t CS  = 1<<PD6; // Serial Clock (active low)
-uint8_t CD  = 1<<PD5; // Serial Data
+uint8_t CLK   = 1<<PD6; // Serial Clock (active low)
+uint8_t DATA  = 1<<PD5; // Serial Data
 
 #define STATE_RELAY   0x00
 #define STATE_COMMAND 0x01
@@ -34,23 +36,27 @@ volatile uint8_t STATE = STATE_RELAY;
 #define WITHOUT_DELAY 0
 
 volatile bool matrix[64];
-volatile bool locked[64];
+volatile uint8_t locked[64];
 volatile uint8_t layout[64];
 
 volatile Serial serial;
+volatile Mapping mappings[16];
 volatile Config* config;
 volatile uint8_t meta;
+volatile bool boot;
+volatile bool reset;
 
 static void (*ResetCrosspointSwitch)(void);
 static void (*StrobeCrosspointSwitch)(void);
 
+static volatile uint8_t usbCommand;
 static volatile uint16_t usbDataReceived;
 static volatile uint16_t usbDataLength;
 static volatile uint8_t *usbData;
 static volatile uint16_t usbDataPos;
 static volatile uint16_t usbDelay;
 
-uint32_t BootKey ATTR_NO_INIT;
+static volatile char version[64];
 
 //------------------------------------------------------------------------------
 
@@ -59,18 +65,13 @@ uint32_t BootKey ATTR_NO_INIT;
 
 //------------------------------------------------------------------------------
 
-void CheckBootloader(void) {
-  if((MCUSR & (1 << WDRF)) && (BootKey == MAGIC)) {
-    BootKey = 0;
-    ((void (*)(void))BOOTLOADER)();
-  }
+void EnterBootloader(void) {
+  eeprom_write_word((uint16_t *)0x0ffe, (uint16_t) 0xb0b0);
+  Reset();
 }
 
-//------------------------------------------------------------------------------
-
-void EnterBootloader(void) {
+void Reset(void) {
   cli();
-  BootKey = MAGIC;
   wdt_enable(WDTO_250MS);
   for(;;);
 }
@@ -94,16 +95,10 @@ void SetupHardware(void) {
 
 //------------------------------------------------------------------------------
 
-/* External Serial control interface 
- *
- * External Hardware may send a 4-bit command nibble followed by an
- * 8-bit argument
- *
- */
-
 void SetupSerial(void) {
   ResetSerial();
-  PCMSK3 = CS;
+  
+  PCMSK3 = CLK;
   PCICR |= (1<<PCIE3);
   sei();
 }
@@ -111,75 +106,149 @@ void SetupSerial(void) {
 //------------------------------------------------------------------------------
 
 void ResetSerial(void) {
-  serial.bit      = 1;
-  serial.data     = 0;
-  serial.command  = 0;
-  serial.argument = 0;
-  serial.expected = SERIAL_COMMAND;
+  ExpectNextSerialByte();
+  serial.command      = 0;
+  serial.arguments[0] = 0;
+  serial.arguments[1] = 0;
+  serial.index        = 0;
 }
 
 //------------------------------------------------------------------------------
 
-void ExpectSerialCommand(void) {
-  ResetSerial();
-}
-
-//------------------------------------------------------------------------------
-
-void ExpectSerialArgument(void) {
+void ExpectNextSerialByte() {
   serial.bit  = 1;
-  serial.data = 0;
-  serial.expected = SERIAL_ARGUMENT;
+  serial.byte = 0;
 }
 
-void ExecuteSerialCommand(uint8_t command, uint8_t argument) {
+//------------------------------------------------------------------------------
 
-  switch(command) {
+void debug(char* fmt, ...) {
+  va_list ap;
+  va_start(ap, fmt);
+
+  char msg[512];
+  vsnprintf(msg, 512, fmt, ap);
+  Type(msg);
+
+  va_end(ap);
+}
+
+//------------------------------------------------------------------------------
+
+void ExecuteSerialCommand() {
+
+  uint8_t port;
+  uint8_t mask;
+  uint8_t key;
+  
+  switch(serial.command) {
 
   case SERIAL_COMMAND_EXECUTE:
-    ExecuteKey(argument);
+    ExecuteKey(serial.arguments[0]);
     break;
 
   case SERIAL_COMMAND_KEY_DOWN:
-    SetCrosspointSwitchLocked(argument, true);
+    SetCrosspointSwitchLocked(serial.arguments[0], true, LOCK_SERIAL);
     PROPAGATE;
     break;
 
   case SERIAL_COMMAND_KEY_UP:
-    SetCrosspointSwitchLocked(argument, false);
+    SetCrosspointSwitchLocked(serial.arguments[0], false, LOCK_SERIAL);
     PROPAGATE;
     break;
 
   case SERIAL_COMMAND_KEY_PRESS:
-    RelayKeyPress(argument);
+    RelayKeyPress(serial.arguments[0]);
     break;
-  }
+
+  case SERIAL_COMMAND_MAP:    
+
+    port = ((serial.arguments[1] & 0x8) == 0) ? 0 : 1;
+    mask = 1<<(serial.arguments[1] & 0x7);
+    key = serial.arguments[0];
+
+    Map(port, mask, key);
+
+    break;
+  }  
 }
 
 //------------------------------------------------------------------------------
 
 ISR(PCINT3_vect) {
 
-  if((PIND & CS) == 0) {
-    if((PIND & CD) != 0) {
-      serial.data |= serial.bit;
+  if((PIND & CLK) != 0) return;
+
+  if((PIND & DATA) != 0) {
+    serial.byte |= serial.bit;
+  }    
+  serial.bit = serial.bit << 1;
+
+  if(serial.bit == 0) { // byte complete
+    
+    if(serial.index == 0) { // received command 
+      serial.command = serial.byte;
+      serial.index = SERIAL_COMMAND_ARITY_FOR[serial.command];
+    }
+    else { // received argument
+      serial.arguments[--serial.index] = serial.byte;
     }    
-    serial.bit = serial.bit << 1;
 
-    if(serial.bit == serial.expected) {
+    if(serial.index == 0) { // all arguments received
+      ExecuteSerialCommand();
+      ResetSerial();
+    }
+    else {
+      ExpectNextSerialByte();
+    }
+  }
+}
 
-      if(serial.expected == SERIAL_COMMAND) {
-        serial.command = serial.data;
-        ExpectSerialArgument();
-        return;
+//------------------------------------------------------------------------------
+
+void SetupMappings(void) {
+  for(uint8_t i=0; i<16; i++) {
+    mappings[i].pins = NULL;
+  }
+}
+
+//------------------------------------------------------------------------------
+
+void Map(uint8_t p, uint8_t mask, uint8_t key) {
+
+  uint8_t volatile *port = (p == PORT_A) ? &PORTB : &PORTC;   
+  uint8_t volatile *pins = (p == PORT_A) ? &PINB : &PINC; 
+  uint8_t volatile *ddr  = (p == PORT_A) ? &DDRB : &DDRC;
+
+  Mapping volatile *mapping = &mappings[mask | (p<<3)];
+
+  mapping->pins = pins;
+  mapping->mask = mask;
+  mapping->key = key;
+  
+  (*ddr) &= ~mask;
+  (*port) |= mask;
+}
+
+//------------------------------------------------------------------------------
+
+void ApplyMappings(void) {
+  bool closed = false;
+  Mapping volatile *mapping;
+  
+  for(uint8_t i=0; i<16; i++) {
+    mapping = &mappings[i];
+    
+    if(mapping->pins == NULL) continue;
+    
+    closed = ((*mapping->pins) & mapping->mask) == 0;      
+    if(closed) {
+      SetCrosspointSwitchLocked(mapping->key, true, LOCK_MAP);
+    }
+    else {
+      if(locked[mappings[i].key] == LOCK_MAP) {
+        SetCrosspointSwitchLocked(mapping->key, false, LOCK_MAP);
       }
-
-      if(serial.expected == SERIAL_ARGUMENT) {
-        serial.argument = serial.data;
-        ExecuteSerialCommand(serial.command, serial.argument);
-        ExpectSerialCommand();
-        return;
-      }      
     }
   }
 }
@@ -201,6 +270,8 @@ void SetupUSB(void) {
   usbData = NULL;
 }
 
+//------------------------------------------------------------------------------
+
 void SetupKeyboardLayout(void) {
   for(uint8_t i=0; i<64; i++) {
     layout[i] = i;
@@ -223,7 +294,23 @@ int ReadEeprom(FILE* file) {
 
 //------------------------------------------------------------------------------
 
-FILE eeprom = FDEV_SETUP_STREAM(NULL, ReadEeprom, _FDEV_SETUP_READ);
+int WriteEeprom(char data, FILE* file) {
+
+  static volatile uint16_t addr = 0;
+  
+  while(EECR & (1<<EEPE));
+  EEAR = addr;
+  EEDR = data;
+  EECR |= (1<<EEMPE);
+  EECR |= (1<<EEPE);
+  addr += 1;
+
+  return 0;
+}
+
+//------------------------------------------------------------------------------
+
+FILE eeprom = FDEV_SETUP_STREAM(WriteEeprom, ReadEeprom, _FDEV_SETUP_RW);
 
 //------------------------------------------------------------------------------
 
@@ -237,7 +324,6 @@ FILE usbdata = FDEV_SETUP_STREAM(NULL, ReadUSBData, _FDEV_SETUP_READ);
 
 //------------------------------------------------------------------------------
 
-
 void ExecuteImmediateCommands(volatile Config* cfg, uint16_t delay) {
   uint16_t ms;
   for(int i=0; i<cfg->num_bindings; i++) {
@@ -308,12 +394,16 @@ void ResetCrosspointSwitch22106(void) {
   PORTD &= ~CPR;
   _delay_ms(1);
   PORTD |= CPR;
+
+  for(uint8_t i=0; i<64; i++) SetCrosspointSwitch(i, false);
 }
 
 void ResetCrosspointSwitch8808(void) {
   PORTD |= CPR;  
   _delay_ms(1);
   PORTD &= ~CPR;
+
+  for(uint8_t i=0; i<64; i++) SetCrosspointSwitch(i, false);
 }
 
 //------------------------------------------------------------------------------
@@ -342,9 +432,11 @@ void SetCrosspointSwitch(uint8_t index, bool closed) {
 
 //------------------------------------------------------------------------------
 
-void SetCrosspointSwitchLocked(uint8_t index, bool closed) {
+void SetCrosspointSwitchLocked(uint8_t index, bool closed, uint8_t lock) {
+  if(locked[index] && locked[index] != lock) return;
+  
   if(!locked[index]) SetCrosspointSwitch(index, closed);
-  locked[index] = closed ? true : false;
+  locked[index] = closed ? lock : LOCK_NONE;
   if(!locked[index]) SetCrosspointSwitch(index, closed);
 }
 
@@ -465,6 +557,18 @@ void ExecuteKey(uint8_t key) {
 
 //------------------------------------------------------------------------------
 
+void ExecuteBindingInteractively(Binding *binding) {
+
+  if(binding->count > 0) {
+    if(--(binding->count) == 0) {
+      ExecuteBinding(binding);
+      Binding_reset(binding);
+    }
+  }
+}
+
+//------------------------------------------------------------------------------
+
 void ExecuteBinding(Binding* binding) {
   Command *command;
   
@@ -594,7 +698,7 @@ void ExecuteCommand(volatile Config *cfg, Command* cmd) {
     break;
 
   case ACTION_BOOT:
-    EnterBootloader();
+    boot = true;
     break;
     
   case ACTION_SWAP:
@@ -624,10 +728,74 @@ void ExecuteCommand(volatile Config *cfg, Command* cmd) {
 
   case ACTION_RESTORE_STATE:
     RestoreState();
-    break;    
+    break;
+
+  case ACTION_MAP:
+    Map(cmd->port, cmd->mask, cmd->data);
+    break;
+
+  case ACTION_SHOW_VERSION:
+    Type((char*)version);
+    Type("\n\n");
+    break;
+
+  case ACTION_SHOW_STATE:
+    ShowState();
+    break;
   }
 }
 
+//------------------------------------------------------------------------------
+
+void SetupVersionString(void) {
+
+  char *v = "firmware " xstr(VERSION) " " __DATE__ " " __TIME__;
+  int len = strlen(v);
+
+  for(uint8_t i=0; i<64; i++) {
+    version[i] = '\0';
+  }
+  uint8_t o=0;
+  
+  for(uint8_t i=0; i<strlen(v); i++) {
+    if(i>1 && v[i-1] == ' ' && v[i] == ' ') continue;
+    version[o++] = tolower(v[i]);
+  }
+}
+
+//------------------------------------------------------------------------------
+
+static char* p2s(char **dst, uint8_t volatile *port, uint8_t volatile *ddr) {
+
+  uint8_t i=0;
+  
+  for(uint8_t bit=0x80; bit; bit>>=1, i++) {
+
+    if((*ddr) & bit) {
+      (*dst)[i] = ((*port) & bit) ? '1' : '0';
+    }
+    else {
+      (*dst)[i] = ((*port) & bit) ? '1' : 'x';
+    }
+  }
+  (*dst)[8] = '\0';
+  return (*dst);
+}
+
+//------------------------------------------------------------------------------
+
+void ShowState(void) {
+  char *state = "00000000";
+
+  Type("a ");
+  Type(p2s(&state, &PORTB, &DDRB));
+  Type("\n");
+
+  Type("b ");
+  Type(p2s(&state, &PORTC, &DDRC)); 
+  Type("\n\n");
+}
+
 //------------------------------------------------------------------------------
  
 void SaveState(void) {
@@ -655,13 +823,15 @@ void RestoreState(void) {
 
 USB_PUBLIC usbMsgLen_t usbFunctionSetup(uint8_t data[8]) {
 
-  usbRequest_t *request = (void*) data;
-
-  switch(request->bRequest) {
+  usbRequest_t *usbRequest = (void*) data;
+  
+  switch(usbRequest->bRequest) {
     
   case KEYMAN64_CTRL:
-    usbDataLength = request->wLength.word;
-    usbDelay = request->wValue.word;
+  case KEYMAN64_FLASH:
+    usbCommand = usbRequest->bRequest;
+    usbDataLength = usbRequest->wLength.word;
+    usbDelay = usbRequest->wValue.word;
     usbDataReceived = 0;
     
     if(usbData == NULL) {
@@ -670,15 +840,27 @@ USB_PUBLIC usbMsgLen_t usbFunctionSetup(uint8_t data[8]) {
     return USB_NO_MSG;
     break;
 
+  case KEYMAN64_BOOT:
+    boot = true;    
+    break;
+    
   case KEYMAN64_STATE:
     usbMsgPtr = (uchar *) config->state;
     return 4;
+    break;
+
+  case KEYMAN64_IDENTIFY:
+    usbMsgPtr = (uchar *) version;
+    return strlen((const char*)version)+1;
+    break;
   }
   return 0;
 }
 
 //------------------------------------------------------------------------------
 
+//------------------------------------------------------------------------------
+
 USB_PUBLIC uchar usbFunctionWrite(uchar *data, uchar len) {
 
   for(int i=0; usbDataReceived < usbDataLength && i < len; i++, usbDataReceived++) {
@@ -689,7 +871,12 @@ USB_PUBLIC uchar usbFunctionWrite(uchar *data, uchar len) {
     return 0;
   }
   else {
-    ExecuteCommandsFromUSBData();
+    if(usbCommand == KEYMAN64_CTRL) {
+      ExecuteCommandsFromUSBData();
+    }
+    else if(usbCommand == KEYMAN64_FLASH) {
+      FlashConfigurationFromUSBData();
+    }
   }
   
   free((void*)usbData);
@@ -712,13 +899,33 @@ void ExecuteCommandsFromUSBData(void) {
 
 //------------------------------------------------------------------------------
 
+void FlashConfigurationFromUSBData(void) {
+  
+  // Restore the saved state prior to overriding it in flash...
+  RestoreState();
+  
+  if(fwrite((void*)usbData, sizeof(uint8_t), usbDataLength, &eeprom) == usbDataLength) {
+
+    // Save the previously restored state back into eeprom
+    SaveState();
+
+    reset = true;
+  }
+}
+
+//------------------------------------------------------------------------------
+
 int main(void) {
 
   SetupHardware();
   SetupKeyboardLayout();
-
-  meta = KEY_ARROWLEFT;
-
+  SetupMappings();
+  SetupVersionString();
+  
+  meta = KEY_BACKARROW;
+  boot = false;
+  reset = false;
+  
   ResetCrosspointSwitch = &ResetCrosspointSwitch8808;
   StrobeCrosspointSwitch = &StrobeCrosspointSwitch8808;  
   
@@ -726,18 +933,26 @@ int main(void) {
 
   Config_read(config, &eeprom) || Config_install_fallback(config);
 
-  ExecuteImmediateCommands(config, WITHOUT_DELAY);
-
   ResetCrosspointSwitch();
+  
+  ExecuteImmediateCommands(config, WITHOUT_DELAY);
 
   SetupUSB();
   
   Binding *binding;
   bool relayMetaKey = true;
-  
+
   while(true) {
 
     usbPoll();
+
+    if(boot || reset) {
+      _delay_ms(250);
+      if(boot) EnterBootloader();
+      if(reset) Reset();
+    }
+
+    ApplyMappings();
     
     switch(STATE) {
 
@@ -746,8 +961,9 @@ int main(void) {
     case STATE_RELAY:
 
       if(QueryKeyDown(meta)) {
+        Config_reset(config);
         relayMetaKey = true;
-        STATE = STATE_COMMAND;	
+        STATE = STATE_COMMAND;
       }
       else {
         RelayMatrix();
@@ -770,7 +986,7 @@ int main(void) {
           
           if(QueryKeyDown(binding->key)) {  
             
-            ExecuteBinding(binding);
+            ExecuteBindingInteractively(binding);
 
             while(!QueryKeyUp(binding->key));
             relayMetaKey = false;
@@ -779,8 +995,7 @@ int main(void) {
       }
       break;
       
-    //========================================
-      
+    //========================================      
     }
   }
   return 0;
