@@ -2,6 +2,7 @@
 #include <stdlib.h>
 #include <stdint.h>
 #include <stdbool.h>
+#include <unistd.h>
 #include <string.h>
 #include <strings.h>
 #include <errno.h>
@@ -35,6 +36,9 @@ char *device;
 uint16_t delay = 0;
 bool preserve_state = false;
 
+DeviceInfo keyman64;
+DeviceInfo usbasp;
+
 //------------------------------------------------------------------------------
 // Utility functions for parsing
 //------------------------------------------------------------------------------
@@ -59,7 +63,13 @@ static uint8_t parseAction(char* str) {
   if(strncasecmp(str, "press",    5) == 0) return ACTION_KEY_PRESS;
   if(strncasecmp(str, "using",    5) == 0) return ACTION_DEFINE_SWITCH;
   if(strncasecmp(str, "save",     4) == 0) return ACTION_SAVE_STATE;
-  if(strncasecmp(str, "restore",  7) == 0) return ACTION_RESTORE_STATE;        
+  if(strncasecmp(str, "restore",  7) == 0) return ACTION_RESTORE_STATE;
+  if(strncasecmp(str, "requires", 8) == 0) return ACTION_REQUIRES;
+  if(strncasecmp(str, "map",      3) == 0) return ACTION_MAP;
+  if(strncasecmp(str, "version",  7) == 0) return ACTION_SHOW_VERSION;
+  if(strncasecmp(str, "state",    5) == 0) return ACTION_SHOW_STATE;
+  if(strncasecmp(str, "status",   6) == 0) return ACTION_SHOW_STATE;      
+  
   return ACTION_NONE;
 }
 
@@ -155,6 +165,8 @@ static bool parseData(char *str, uint8_t *data) {
   return result;    
 }
 
+//------------------------------------------------------------------------------
+
 static bool parsePolicy(Command* command, char *str) {
   bool result = true;
 
@@ -249,6 +261,99 @@ static bool isSymbol(char *name) {
     return false;
 }
 
+//------------------------------------------------------------------------------
+
+static void unescape(char *str) {
+  char *saved_str = str;
+  char *input = strdup(str);
+  char *saved_input = input;
+  int closing;
+  char in, out;
+  
+  for(input; input[0] != '\0'; input++) {
+    char hex[] = { 0, 0, 0 };
+
+    in = input[0];
+    if(in == '\\') {
+      switch(input[1]) {
+      case '\\':
+        out = '\\';
+        break;        
+      case 'r':
+        out = '\r';
+        break;
+      case 'n':
+        out = '\n';
+        break;
+      case 'f':
+        out = '\f';
+        break;
+      case 'x':        
+        hex[0] = input[2];
+        hex[1] = input[3];
+        out = strtol(hex, NULL, 16) & 0xff; 
+        input+=2;
+        break;
+      case '{':
+        closing = strcspn(input+2, "}");
+        input[2+closing] = '\0';
+        out = strtol(input+2, NULL, 10);
+        input += closing + 1;
+        break;
+      default:
+        out = in;
+        fprintf(stderr, "Warning: \\%c: undefined escape sequence in string literal (skipped)\n", input[1]);        
+        input++;
+        continue;
+      }              
+      str[0] = out;
+      input++;
+    }
+    else {
+      out = in;
+      str[0] = out;
+    }
+    str++;
+  }
+  free(saved_input);
+
+  str[0] = '\0';
+  str = saved_str;
+}
+
+//------------------------------------------------------------------------------
+
+static char* escape(char* str) {
+  char* output = (char*) calloc(strlen(str)*6, sizeof(char));
+  char seq[7];
+  
+  for(int i=0; i<strlen(str); i++) {    
+    switch(str[i]) {
+    case '\\':
+      strcat(output, "\\\\");
+      break;
+    case '\n':
+      strcat(output, "\\n");
+      break;
+    case '\r':
+      strcat(output, "\\r");
+      break;
+    case '\f':
+      strcat(output, "\\f");
+      break;
+    default:
+      if((str[i] >= 32  && str[i] <= 127) || (str[i] >= 192 && str[i] <= 223)) {
+        output[strlen(output)] = str[i];
+      }
+      else {
+        sprintf(seq, "\\{%d}", (unsigned char)str[i]);
+        strcat(output, seq);
+      }
+    }
+  }
+  return output;
+}
+
 //------------------------------------------------------------------------------
 // Functions for parsing config files
 //------------------------------------------------------------------------------
@@ -265,7 +370,8 @@ bool Config_parse(Config* self, FILE* in) {
   char *comment;
   int pos = 0;
   int trailing = 0;
-  
+  uint8_t times = 0;
+
   Binding* binding;
   Command* command;
   uint8_t key;
@@ -341,30 +447,24 @@ bool Config_parse(Config* self, FILE* in) {
 
     // skip leading whitespace of command spec
     line += strspn(line, ws);
-
+    
     // try to parse command...        
     command = Command_new();
-
+    
     if(!Command_parse(command, line)) {
       fprintf(stderr, "error: line %d: '%s': invalid command specification\n", pos, line);
       free(command);
       return false;
-    }
+    }    
 
-    // get existing binding or add new binding...
-    if(!Config_has_binding(self, key)) {
-      binding = Config_get_binding(self, key);
+    if(command->action == ACTION_REQUIRES) {
+      Config_get_or_create_binding(self, key)->times = command->data;
     }
-    else {
-      binding = Binding_new();
-      Binding_set_key(binding, key);
-      Config_add_binding(self, binding);
+    else {    
+      Binding_add(Config_get_or_create_binding(self, key), command);
     }
-
-    // append command to binding
-    Binding_add(binding, command);
   }
-
+  
   free(buffer);
   return true;
 }
@@ -425,11 +525,12 @@ bool Command_parse(Command* self, char* spec) {
   bool error = false;
   StringList* words = StringList_new();
   char* str;
+  char* failed;
   uint8_t data;
   uint16_t index;
   uint32_t value;
   bool has_policy = false;
-  int i = 0;
+  int i = 0;  
   
   StringList_append_tokenized(words, spec, ws);
 
@@ -443,6 +544,16 @@ bool Command_parse(Command* self, char* spec) {
     goto error;
   }
 
+  if(self->action == ACTION_REQUIRES) {
+    str = StringList_get(words, i++);
+    self->data = strtol(str, &failed, 0);
+    if(str == failed) {
+      fprintf(stderr, "error: missing times for requires\n");
+      goto error;
+    }
+    goto done;
+  }
+  
   if(self->action == ACTION_TYPE) {
     str = spec + strcspn(spec, ws) + 1;
 
@@ -450,6 +561,8 @@ bool Command_parse(Command* self, char* spec) {
       str = str + strcspn(str, ws) + 1;
     }
 
+    unescape(str);
+    
     if(!Config_has_string(config, str, &index)) {
       index = Config_add_string(config, str);
     }
@@ -613,6 +726,7 @@ void Config_write(Config *self, FILE *out) {
 
 void Binding_write(Binding *self, FILE* out) {
   Key_write(self->key, out);
+  fputc(self->times, out);
   fputc(self->num_commands, out);
   
   for(int i=0; i<self->num_commands; i++) {
@@ -651,8 +765,8 @@ void State_write(State *self, FILE* out) {
 bool State_fetch(State *self) {
   int result;
   
-  if((result = usb_receive(device, KEYMAN64_STATE, 0, (uint8_t*) self, sizeof(State))) < 0) {
-    fprintf(stderr, "error: could send usb control message: %s\n", libusb_strerror(result));
+  if((result = usb_receive(&keyman64, KEYMAN64_STATE, 0, 0, (uint8_t*) self, sizeof(State))) < 0) {
+    fprintf(stderr, "error: could not send usb control message: %s\n", libusb_strerror(result));
   }
   return result;
 }
@@ -670,6 +784,13 @@ void Config_print(Config *self, FILE* out) {
 //------------------------------------------------------------------------------
 
 void Binding_print(Binding *self, FILE* out) {
+
+  if(self->times) {
+    if(self->key != KEY_IMMEDIATE) {
+      Key_print(self->key, out);
+    }
+    fprintf(out, "requires %dx\n", self->times);
+  }
   
   for(int i=0; i<self->num_commands; i++) {
     if(self->key != KEY_IMMEDIATE) {
@@ -711,6 +832,10 @@ void Command_print(Command *self, FILE* out) {
   case ACTION_DEFINE_SWITCH: action = "using";    break;
   case ACTION_SAVE_STATE:    action = "save";     break;
   case ACTION_RESTORE_STATE: action = "restore";  break;                                
+  case ACTION_REQUIRES:      action = "requires"; break;
+  case ACTION_MAP:           action = "map";      break;
+  case ACTION_SHOW_VERSION:  action = "version";  break;
+  case ACTION_SHOW_STATE:    action = "state";    break;        
   };
 
   if(self->policy == POLICY_EVEN) {
@@ -723,8 +848,9 @@ void Command_print(Command *self, FILE* out) {
   
   if(self->action == ACTION_TYPE) {
     uint16_t index = self->mask | (self->data << 8);
-    fprintf(out, "%s ", action);
-    fprintf(out, "%s\n", config->strings[index]);
+    char *escaped = escape(config->strings[index]);
+    fprintf(out, "%s %s\n", action, escaped);
+    free(escaped);
     return;
   }
   uint8_t mask  = 0;
@@ -797,7 +923,7 @@ void Command_print(Command *self, FILE* out) {
               end);
     }
   }
-  if(self->action == ACTION_SET) {
+  if(self->action == ACTION_SET || self->action == ACTION_MAP) {
     fprintf(out, "$%02X", self->data);
   }
   fprintf(out, "\n");
@@ -832,6 +958,22 @@ void fmemupdate(FILE *fp, void *buf,  uint16_t size) {
 
 //------------------------------------------------------------------------------
 
+static void prepare_devices(void) {
+  strncpy(keyman64.path, device, 4096);
+  keyman64.vid = KEYMAN64_VID;
+  keyman64.pid = KEYMAN64_PID;
+
+#if linux
+  strncpy(usbasp.path, "/dev/usbasp", 4096);
+#else
+  strncpy(usbasp.path, "usbasp", 4096);
+#endif
+  usbasp.vid = USBASP_VID;
+  usbasp.pid = USBASP_PID;
+}
+
+//------------------------------------------------------------------------------
+
 int main(int argc, char **argv) {
 
 #if linux
@@ -841,6 +983,7 @@ int main(int argc, char **argv) {
 #endif
 
   int result = EXIT_SUCCESS;
+  usb_quiet = false;
   
   device = (char*) calloc(strlen(default_device)+1, sizeof(char));
   strcpy(device, default_device);
@@ -852,13 +995,14 @@ int main(int argc, char **argv) {
     { "keys",     no_argument,       0, 'k' },
     { "delay",    required_argument, 0, 'D' },
     { "preserve", no_argument,       0, 'p' },
+    { "identify", no_argument,       0, 'i' },
     { 0, 0, 0, 0 },
   };
   int option, option_index;
 
 
   while(1) {
-    option = getopt_long(argc, argv, "hvd:kD:p", options, &option_index);
+    option = getopt_long(argc, argv, "hvd:kD:pi", options, &option_index);
 
     if(option == -1)
       break;
@@ -892,6 +1036,11 @@ int main(int argc, char **argv) {
     case 'p':
       preserve_state = true;
       break;      
+
+    case 'i':
+      identify();
+      goto done;
+      break;
       
     case '?':
     case ':':
@@ -899,12 +1048,20 @@ int main(int argc, char **argv) {
     }    
   }
 
+  prepare_devices();
+  
   argc -= optind;
   argv += optind;
 
   if(argc && (strcmp(argv[0], "convert") == 0)) {
     result = convert(--argc, ++argv);
   }
+  else if(argc && (strcmp(argv[0], "configure") == 0)) {
+    result = configure(--argc, ++argv);
+  }
+  else if(argc && (strcmp(argv[0], "update") == 0)) {
+    result = update(--argc, ++argv);
+  }
   else {
     result = command(argc, argv);
   }
@@ -983,7 +1140,7 @@ int command(int argc, char **argv) {
   fmemupdate(out, data, size);  
   fclose(out);
 
-  if((result = usb_send(device, KEYMAN64_CTRL, delay, data, size)) < 0) {
+  if((result = usb_send(&keyman64, KEYMAN64_CTRL, delay, 0, data, size)) < 0) {
     fprintf(stderr, "error: could send usb control message: %s\n", libusb_strerror(result));
     goto done;
   }
@@ -1065,10 +1222,161 @@ int convert(int argc, char **argv) {
 
 //------------------------------------------------------------------------------
 
+int configure(int argc, char **argv) {
+
+  int result = EXIT_FAILURE;
+
+  FILE *in  = stdin;
+  FILE *out = NULL;
+  uint8_t *data = NULL;
+  uint16_t size = 0;
+
+  config = Config_new();
+  
+  if(argc >= 1 && (strncmp(argv[0], "-", 1) != 0)) {
+
+    if((in = fopen(argv[0], "rb")) == NULL) {
+      fprintf(stderr, "%s: %s\n", argv[0], strerror(errno));
+      goto done;
+    }
+  }
+
+  if(in == stdin) {
+    fprintf(stderr, "reading from stdin...\n");
+  }
+
+  if(Config_read(config, in) || Config_parse(config, in)) {
+
+    if(preserve_state) {
+      if(!State_fetch(config->state)) {
+        fprintf(stderr, "warning: could not fetch saved state from device\n");
+      }        
+    }
+
+    data = (uint8_t*) calloc(4096, sizeof(char));
+  
+    if((out = fmemopen(data, 4096, "wb")) == NULL) {
+      fprintf(stderr, "error: %s\n", strerror(errno));
+      goto done;
+    }
+  
+    Config_write(config, out);
+    size = ftell(out);
+    fmemupdate(out, data, size);  
+    fclose(out);
+
+    if(!usb_ping(&keyman64)) {
+      fprintf(stderr, "error: could not connect to keyman64\n");
+      goto done;
+    }
+
+    fprintf(stderr, "Flashing configuration: %d bytes...", size);
+    fflush(stderr);
+
+    result = usb_send(&keyman64, KEYMAN64_FLASH, 0, 0, data, size);
+
+    fprintf(stderr, (result == size) ? "ok\n" : "failed!\n");
+  }
+
+ done:
+  Config_free(config);
+  fclose(in);
+
+  return result;
+}
+
+//------------------------------------------------------------------------------
+
+int update(int argc, char **argv) {
+  int result = EXIT_FAILURE;
+  FILE *in;
+  uint8_t *data = NULL;
+  uint16_t size = 0;
+  
+  struct stat st;
+  
+  if(!argc) {
+    usage();
+    goto done;
+  }
+  
+  if((in = fopen(argv[0], "rb")) == NULL) {
+    goto error;
+  }
+
+  if(fstat(fileno(in), &st) == -1) {
+    goto error;
+  }
+
+  size = st.st_size;
+  data = (uint8_t *) calloc(size, sizeof(uint8_t));
+
+  if(fread(data, sizeof(uint8_t), size, in) != size) {
+    goto error;
+  }
+  fclose(in);
+
+  if(usb_ping(&keyman64)) {
+    fprintf(stderr, "Entering bootloader"); fflush(stderr);
+    usb_control(&keyman64, KEYMAN64_BOOT);
+
+    uint8_t tries = 10;
+    while(!usb_ping(&usbasp)) {
+      fprintf(stderr, "."); fflush(stderr);
+
+#if windows 
+      Sleep(1000);
+#else
+      sleep(1);
+#endif
+      
+      if(!--tries) break;
+    }
+  }
+
+  if(usb_ping(&usbasp)) {
+    usb_control(&usbasp, USBASP_CONNECT);
+  
+    for(uint32_t i=0; i<size+64; i+=64) {
+      usb_send(&usbasp, USBASP_WRITEFLASH,
+               (uint16_t) (i & 0xffff), (uint16_t) (i>>16),
+               data+i, 64);
+      fprintf(stderr, "\rUpdating application: %d of %d bytes transferred...",
+              (i<size) ? i : size , size);
+    }
+    fprintf(stderr, "ok\n");
+
+    usb_quiet = true;  
+    usb_control(&usbasp, USBASP_DISCONNECT);
+  }
+  else {
+    fprintf(stderr, "error: could not connect to usbasp bootloader\n");
+  }
+  
+ done:
+  if(data != NULL) free(data);
+  return result;
+
+ error:
+  fprintf(stderr, "%s: %s\n", argv[0], strerror(errno));
+  goto done;
+}
+
+//------------------------------------------------------------------------------
+
+void identify(void) {
+  char id[64];
+  if(usb_receive(&keyman64, KEYMAN64_IDENTIFY, 0, 0, (uint8_t*) id, 64) > 0) {
+    printf("%s\n", id);
+  }
+}
+
+//------------------------------------------------------------------------------
+
 void version(void) {
-  printf("keyman64 version %.1f\n", VERSION);
-  printf("Copyright (C) 2015 Henning Bekel.\n");
-  printf("License GPLv2: GNU GPL version 2 <http://gnu.org/licenses/gpl.html>.\n");
+  printf("Keyman64 v%.1f\n", VERSION);    
+  printf("Copyright (C) 2016 Henning Bekel.\n");
+  printf("License GPLv3: GNU GPL version 3 <http://gnu.org/licenses/gpl.html>.\n");
   printf("This is free software: you are free to change and redistribute it.\n");
   printf("There is NO WARRANTY, to the extent permitted by law.\n");
 }
@@ -1081,8 +1389,10 @@ void usage(void) {
   printf("Usage:\n");
   printf("      keyman64 <options>\n");
   printf("      keyman64 <options> convert [<infile>|-] [<outfile>|-]\n");
+  printf("      keyman64 <options> configure [<infile>|-]\n");
+  printf("      keyman64 <options> update <firmware>\n");    
   printf("      keyman64 [<options>] <command>\n");
-  printf("      keyman64 [<options>] [<file>]\n");    
+  printf("      keyman64 [<options>] [<script>]\n");    
   printf("\n");
   printf("  Options:\n");
   printf("           -v, --version  : print version information\n");
@@ -1094,16 +1404,19 @@ void usage(void) {
 #endif
   printf("           -D, --delay    : delay in ms between commands\n");  
   printf("           -k, --keys     : list key names and synonyms\n");
-  printf("           -p, --preserve : preserve saved state during convert\n");  
+  printf("           -p, --preserve : preserve saved state during convert\n");
+  printf("           -i, --identify : request firmware identification via USB\n");
   printf("\n");
-  printf("  Conversion arguments:\n");
-  printf("           <infile>  : input file, format is autodetected\n");
-  printf("           <outfile> : output file, format determined by extension:\n");
+  printf("  Files:\n");
+  printf("           <infile>   : input file, format is autodetected\n");
+  printf("           <outfile>  : output file, format determined by extension\n");
+  printf("           <script>   : script file containing keyman64 commands\n");
+  printf("           <firmware> : binary firmware image\n");    
   printf("\n");
   printf("           *.conf : plain text config file format\n");
   printf("           *.bin  : binary file format (default)\n");  
   printf("\n");
-  printf("           Missing arguments default to stdin or stdout\n");
+  printf("           Optional arguments default to stdin or stdout\n");
   printf("\n");
   printf("  Command:\n");
   printf("            (any valid keyman64 command)\n");
@@ -1118,7 +1431,7 @@ void keys(void) {
   int total = sizeof(symbols)/sizeof(Symbol);
 
   for(int i=0; i<64; i++) {
-    printf("%02X: ", i);
+    printf("$%02X %2d   ", i, i);
     symbol = symbols[i];
     printf("%s", symbol.name);
     for(int k=64; k<total; k++) {
@@ -1132,3 +1445,4 @@ void keys(void) {
 }
 
 //------------------------------------------------------------------------------
+
