@@ -1,9 +1,10 @@
 /* -*- mode: kasm -*- */
 
 .namespace Command {
+.label none  = $00
 .label exec  = $01
 .label down  = $02
 .label up    = $03
 .label press = $04
-.label input = $05
+.label map   = $05
 }
